package soal.ppr

import scala.collection._
import soal.util.DiscreteAliasSampler

/*
 * Created by plofgren on 4/17/15.
 */

trait TargetSetPrecomputation {
  // Maps v to sum_t reverseVector_t(v)
  def reverseVectorSum(v: Int): Float
  // Maps v to (sampler returning t with probability proportional to r_t(v))
  def sampleTargetFrom(v: Int): Int
  def size: Int
}
case class SimpleTargetSetPrecomputation(reverseVectorSumMap: Map[Int, Float],
                                          targetSamplerMap: Map[Int, DiscreteAliasSampler]
  ) extends TargetSetPrecomputation {
  def reverseVectorSum(v: Int) = reverseVectorSumMap(v)
  def sampleTargetFrom(v: Int) = {
    assert(targetSamplerMap.contains(v), s"Error: trying to sample target from intermediate node $v using targetSamplerMap " +
      s"of size ${targetSamplerMap.size}.")
    targetSamplerMap(v).sample()
  }
  def size: Int =
    reverseVectorSumMap.size + (targetSamplerMap.values map { _.size }).sum
}
