package soal.ppr.experiments

import java.io._
import scala.util.Random
import java.nio.channels.FileChannel
import java.nio.file.StandardOpenOption
import java.nio.channels.FileChannel.MapMode
import soal.markov.BidirectionalTransitionEstimator
import scala.io.Source

object Main {
  def main(args: Array[String]) {
    println("args: " + args.toSeq)
    assert(args.size > 2, "Arguments needed")
    lazy val graphFilename = args(1)
    lazy val graph = ExperimentUtils.readGraph(graphFilename) // Lazy evaluation to save time if
    // other parameters are invalid
    // alternatively, args(1) could be file with graph names
    lazy val graphNames = Source.fromFile(graphFilename).getLines().toArray
    val teleportProbability = 0.2f
    args.head match {
      case "ppr-runtime" =>
        val pairCount = args(2).toInt
        val msPerWalks = (args drop 3) map (_.toFloat)
        new RuntimeExperiments(graph, teleportProbability).forwardReverseTimesExperiment(pairCount, msPerWalks)
      /*case "runtime_balanced" => RuntimeExperiments.measureRuntime(graph, 100, true)
      case "accuracy" => AccuracyExperiments.measureRelativeError(graph, 10, 100)
      case "ppr_ranking" =>  {
        assert(args.size == 5)
        val idRemappingPath = args(2)
        val startTwitterId = args(3).toInt
        val walkCount = args(4).toInt
        RankingPreliminaryExperiment.topTargetsForUsers(graph, idRemappingPath, startTwitterId, walkCount)
      }
      case "ranking_csv_expand" => {
        assert(args.size == 2)
        val rankingPath = args(1)
        RankingPreliminaryExperiment.csvToExpandedHTML(rankingPath)
      }
      case "csv_experiment" => {
        assert(args.size == 4)
        val csvFilePathPrefix = args(2)
        val nodeCount = args(3).toInt
        CSVExportExperiment.createCSVFiles(graph, csvFilePathPrefix, nodeCount)
      }
      case "ff_size" => {
        assert(args.size == 3)
        //val maxResidual = args(2).toFloat
        val nodeCount = args(2).toInt
        //ForwardFrontierExperiment.forwardFrontierSizes(graph, maxResidual, nodeCount)
        ForwardFrontierExperiment.forwardFrontierSizesVaryingResidual(graph, nodeCount)
      }*/
      case "storage" => {
        assert(args.size == 3)
        val nodeCount = args(2).toInt
        StorageExperiments.storageSizesVaryingResidual(graph, teleportProbability, nodeCount)
      }
/*
      case "search-accuracy" => {
        assert(args.size == 4)
        val targetSetSize = args(2).toInt
        val trialCount = args(3).toInt
        PPRSearchExperiments.approximateSamplingAccuracyExperiment(
          graph, teleportProbability, targetSetSize, trialCount)
      }
*/
      case "search-runtime" => {
        assert(args.size >= 4)
        val trialCount = args(2).toInt
        val targetSetSizes = (args drop 3) map (_.toInt)
        PPRSearchExperiments.runtimeAndAccuracyExperiment(graph, teleportProbability, trialCount, targetSetSizes)
      }

      case "cross-ppr-runtime" => {
        assert(args.size == 3)
        val pairCount = args(2).toInt
        val pageRanks = ExperimentUtils.loadOrComputePageRank(graph, graphFilename, teleportProbability)
        println("global PageRank sum " + pageRanks.sum)
        val experiment = new CrossPartitioningExperiments(graph, teleportProbability, pageRanks)
        experiment.shardedRuntimeExperiment(pairCount)
      }

      case "heat-accuracy" =>
        assert(args.size == 3)
        val trialCount = args(2).toInt
        HeatKernelAccuracyExperiments.accuracyExperiment(graph, trialCount)

      case "heat-accuracies" =>
        assert(args.size == 3)
        val trialCount = args(2).toInt
        HeatKernelAccuracyExperiments.measureAccuracyOnGraphs(graphNames, trialCount)

      case "heat-runtime" =>
        assert(args.size == 3)
        val trialCount = args(2).toInt
        val meanTimes = HeatKernelRuntimeExperiments.measureRuntime(graph, trialCount)
        println(s"meanTimes $meanTimes")

      case "heat-runtimes" =>
        assert(args.size == 3)
        val trialCount = args(2).toInt
        HeatKernelRuntimeExperiments.measureRuntimeOnGraphs(graphNames, trialCount)

      case "sharded-ppr-accuracy" =>
        assert(args.size == 4)
        val pairCount = args(2).toInt
        val trialCount = args(3).toInt
        new ShardedPageRankExperiments(graph, teleportProbability).run(pairCount, trialCount)

      case "residuals" =>
        assert(args.size == 3)
        val pairCount = args(2).toInt
        new ReversePushExperiments(graph, teleportProbability).residualDistributionExperiment(pairCount)

      case "path-prob" => {
        assert(args.size == 3)
        val pairCount = args(2).toInt
        val pageRanks = ExperimentUtils.loadOrComputePageRank(graph, graphFilename, teleportProbability)
        println("global PageRank sum " + pageRanks.sum)
        val experiment = new PathLengthExperiments(graph, teleportProbability, pageRanks)
        experiment.pathLengthProbExperiment(pairCount)
      }

      case "mem-test" =>
        memTest()

      case "node-degree" =>
        val nodeId = args(2).toInt //23933986
        val node = graph.getNodeById(nodeId).get
        println(s"node $nodeId has outDegree " + node.outboundCount + " and inDegree " + node.inboundCount)

      case "undirected-accuracy" =>
        assert(args.size == 3)
        val trialCount = args(2).toInt
        new UndirectedPPRExperiments(graph, teleportProbability).accuracyExperiment(trialCount)
      case "undirected-runtime" =>
        val pairCount = args(2).toInt
        val msPerWalks = (args drop 3) map (_.toFloat)
        new UndirectedPPRExperiments(graph, teleportProbability).walkPushTimesTimesExperiment(pairCount,
          msPerWalks)

      case "export-search-matrix" =>
        assert(args.size >= 5)
        val graphName = args(2)
        val trialCount = args(3).toInt
        val targetSetSizes = (args drop 4) map (_.toInt)
        PPRSearchExperiments.exportReverseMatrixExperiment(graph, graphName, teleportProbability,
          trialCount, targetSetSizes)

      case otherString => println("Unrecognized experiment " + otherString)
    }
  }

  def memTest() {
    val intCount = 250// * 1000 * 1000
    val accessCount = 100
    val file = new File("randomData")
    val in = new DataInputStream(new BufferedInputStream(new FileInputStream(file)))
    val dataArray = new Array[Int](intCount)
    for (i <- 0 until intCount) {
      dataArray(i) = in.readInt()
    }
    in.close()
    println("done reading data")
    val random = new Random()
    val sum1 = ExperimentUtils.printTime("array access time:") {
      ((0 until accessCount) map { j =>
        dataArray(random.nextInt(intCount))
      }).sum
    }


    println("memory mapping file")
    val fileChannel = FileChannel.open(file.toPath, StandardOpenOption.READ)
    val mappedFile = fileChannel.map(MapMode.READ_ONLY, 0, intCount * 4)
    mappedFile.load()
    val sum2 = ExperimentUtils.printTime("mapped file access time:") {
      ((0 until accessCount) map { j =>
        mappedFile.getInt(random.nextInt(intCount))
      }).sum
    }
    fileChannel.close()

    val randomAccess = new RandomAccessFile(file, "r")
    val sum3 = ExperimentUtils.printTime("random file access time:") {
      ((0 until accessCount) map { j =>
        randomAccess.seek(random.nextInt(intCount) * 4)
        randomAccess.readInt()
      }).sum
    }
    println(s"$sum1 $sum2 $sum3")
  }
}
