package soal.ppr.experiments

import java.io.{BufferedWriter, FileWriter, File}

import soal.util.{DiscreteAliasSampler, CollectionsUtil}
import com.twitter.cassovary.graph.{Node, DirectedGraph}
import java.util.Random
import scala.collection.mutable
import scala.collection.mutable.ArrayBuffer
import scala.collection.Map
import soal.ppr.{BidirectionalPPRPrecomputor, BidirectionalPPRSearcher, BidirectionalPPREstimator}

/*
 * Created by plofgren on 4/21/15.
 */

class PPRSearchExperiments(graph: DirectedGraph[Node], teleportProb: Float) {
  val random = new Random()
  val maxK = 10
  val pprEstimator = new BidirectionalPPREstimator(graph, teleportProb, random)
  val pprSearcher = new BidirectionalPPRSearcher(graph, teleportProb)
  val pprPrecomputor = new BidirectionalPPRPrecomputor(graph, teleportProb, random)

  /**
   * Returns a random node with at least one outneighbor.
   */
  def randomNode(): Node = {
    val resultOption = graph.getNodeById(random.nextInt(graph.maxNodeId))
    resultOption.getOrElse(randomNode())
  }

  def randomNonIsolatedStart(): Int = {
    var start = randomNode()
    while (start.outboundCount < 1) {
      println(s"rejecting start ${start.id} with outdegree ${start.outboundCount} because it " +
        s"has insuffucient outDegree")
      start = randomNode()
    }
    start.id
  }
/*
  def measureAccuracyToTargets(
      targets: Seq[Int],
      minPPR: Float,
      reverseThreshold: Float,
      walkCounts: Seq[Int],
      startNodeCount: Int
      ): AccuracyResult = {
    println(s"# Precomputing for targets of size ${targets.size}...")
    val (precomputation, precomputationTime) = ExperimentUtils.time {
      pprSearcher.precomputeForTargetSet(targets, reverseThreshold)
    }
    println(s"# Precomputing took $precomputationTime seconds")
    println(s"# Precomputation size: ${precomputation.size * 12 / 1024.0} KB")

    val targetSet = new mutable.HashSet[Int] ++= targets
    val samplingPrecisionAtKSums = ArrayBuffer.fill(walkCounts.size)(ArrayBuffer.fill(maxK + 1)(0.0))
    val biPPRPrecisionAtKSums = ArrayBuffer.fill(walkCounts.size)(ArrayBuffer.fill(maxK + 1)(0.0))
    val mcPrecisionAtKSums =
      ArrayBuffer.fill(walkCounts.size)(ArrayBuffer.fill(maxK + 1)(0.0))
    val samplingTotalPPRAtKSums =  ArrayBuffer.fill(walkCounts.size)(ArrayBuffer.fill(maxK + 1)(0.0))
    val totalTargetPPRs =  new ArrayBuffer[Double]()

    for (startIndex <- 0 until startNodeCount) {
      // Find a start node with significant PPR to targets
      val start = randomNonIsolatedStart()
      val trueTargetPPRs: collection.Map[Int, Float] = estimateTrueTargetPPRs(start, targets)
      totalTargetPPRs += trueTargetPPRs.values.sum
//      // Find a start node with significant PPR to targets
//      var start: Node = null
//      var trueTargetPPRs: collection.Map[Int, Float] = Map.empty
//      while (trueTargetPPRs.values.sum < 10.0 / graph.nodeCount) {
//        if (start != null)
//          println(s"rejecting start ${start.id} with outdegree ${start.outboundCount} because it " +
//            s"has insuffucient total PPR to targets")
//        start = randomNode()
//        trueTargetPPRs = estimateTargetPPRs(start.id, targets)
//      }

      for ((walkCount, walkCountI) <- walkCounts.zipWithIndex) {
        val accuracyMultiplier = walkCount / walkCounts(0)
        println(s"measuring accuracy of each method using $walkCount walks (multiplier $accuracyMultiplier)")
        val samplingTargetWeights = ExperimentUtils.printTime("samplingSearch took ") {
          pprSearcher.sampleTargets(start, precomputation, walkCount, walkCount)
        }

        val (samplingPrecisions, samplingTotalPPRs) = measurePrecisionAndTotalPPRAtKs(
          targets, samplingTargetWeights, trueTargetPPRs, maxK)
        CollectionsUtil.plusEquals(samplingPrecisionAtKSums(walkCountI), samplingPrecisions)

        val relativeError = 0.1f / accuracyMultiplier
        val biPPRTargetWeights = ExperimentUtils.printTime("biPPRSearch took ") {
          (targets map {
            t => (t, pprEstimator.estimatePPR(start, t, minPPR, relativeError))
          }).toMap
        }
        val (biPPRPrecisions, biPPRTotalPPRs) = measurePrecisionAndTotalPPRAtKs(
          targets, biPPRTargetWeights, trueTargetPPRs, maxK)
        CollectionsUtil.plusEquals(biPPRPrecisionAtKSums(walkCountI), biPPRPrecisions)
        CollectionsUtil.plusEquals(samplingTotalPPRAtKSums(walkCountI), biPPRTotalPPRs)


        val relevantSampleRequirement = accuracyMultiplier * 20
        val mcTargetWeights = ExperimentUtils.printTime("mcSearch took ") {
          estimateMontoCarloTargetWeightsDynamic(start, targetSet, relevantSampleRequirement)
        }
        val (mcPrecisions, mcTotalPPRs) = measurePrecisionAndTotalPPRAtKs(
          targets, mcTargetWeights, trueTargetPPRs, maxK)
        CollectionsUtil.plusEquals(mcPrecisionAtKSums(walkCountI), mcPrecisions)
      }
    }
    AccuracyResult(
      CollectionsUtil.scale(biPPRPrecisionAtKSums, 1.0 / startNodeCount),
      CollectionsUtil.scale(samplingPrecisionAtKSums, 1.0 / startNodeCount),
      CollectionsUtil.scale(mcPrecisionAtKSums, 1.0 / startNodeCount),
      CollectionsUtil.scale(samplingTotalPPRAtKSums, 1.0 / startNodeCount),
      totalTargetPPRs
    )
  }

    /*if (walkCountI == 0 && k == 3) {
  // For conciseness, just record for first iteration
  val totalTargetPPR = trueTargetPPRs.values.sum
  println(s"totalTargetPPR $totalTargetPPR    precision@3 " +
    topK.count(trueTopK.contains))
  totalTargetPPRs += totalTargetPPR
  precisionAt3s += topK.count(trueTopK.contains)
}*/
*/
  def measurePrecisionAndTotalPPRAtKs(targets: Seq[Int],
                         targetWeights: collection.Map[Int, Float],
                         trueTargetPPRs: collection.Map[Int, Float],
                         maxK: Int): (IndexedSeq[Double], IndexedSeq[Double]) = {
    val sortedTargets = (targets sortBy targetWeights).reverse
    val trueSortedTargets: Seq[Int] = (targets sortBy trueTargetPPRs).reverse
    ((1 to maxK) map { k: Int =>
      val trueTopK = (trueSortedTargets take k).toSet
      val topK = sortedTargets take k
      val precision = topK.count(trueTopK.contains).toDouble / k
      val totalPPRAtK = (topK map trueTargetPPRs).sum
      val trueTotalPPRAtK = (trueTopK map trueTargetPPRs).sum
      (precision, totalPPRAtK.toDouble / trueTotalPPRAtK)
    }).unzip
  }

  def estimateTrueTargetPPRs(start: Int, targets: Seq[Int]): collection.Map[Int, Float] = {
    val minimumPPR = 1.0f / graph.nodeCount
    val relativeError = 0.01f // Use high accuracy-- this should give mean relative error
    // less than 0.01.
    def estimate(t: Int) = pprEstimator.estimatePPR(start, t, minimumPPR, relativeError)
    val (result, time) = ExperimentUtils.time {
      mutable.HashMap() ++= (targets map { t => (t, estimate(t))})
    }
    println(f"computing ground truth took $time%.2g s")
    result.withDefaultValue(0.0f)
  }

  def estimateMontoCarloTargetWeightsDynamic(
      s: Int, targets: collection.Set[Int],
      relevantSampleRequirement: Int
      ): collection.Map[Int, Float] = {
    val maxSampleCount = 1000 * 1000

    val targetWeights = CollectionsUtil.efficientIntFloatMapWithDefault0()
    var relevantSampleCount = 0
    var sampleCount = 0
    while (relevantSampleCount < relevantSampleRequirement && sampleCount < maxSampleCount) {
      val t = pprEstimator.samplePPR(s)
      if (targets.contains(t)) {
        targetWeights(t) += 1
        relevantSampleCount += 1
      }
      sampleCount += 1
    }
    println(s"mc did $sampleCount walks to find $relevantSampleCount relevant targets")
    if (relevantSampleCount < relevantSampleRequirement)
      println(s"Requirement not met! ----------------- required $relevantSampleRequirement")
    targetWeights
  }

  def estimateMontoCarloTargetWeights(
                                       s: Int, targets: collection.Set[Int],
                                       walkCount: Int
                                       ): collection.Map[Int, Float] = {
    val targetWeights = CollectionsUtil.efficientIntFloatMapWithDefault0()
    var relevantSampleCount = 0
    for (i <- 0 until walkCount) {
      val t = pprEstimator.samplePPR(s)
      if (targets.contains(t)) {
        targetWeights(t) += 1
      }
    }
    targetWeights
  }
  /*
  def estimateTrueTargetPPRs(start: Int, targets: Seq[Int], reverseThreshold: Float,
      precomputation: TargetSetPrecomputation):
      collection.Map[Int, Float] = {
    val chernoffMultiplier: Int = 1400  // I think (?) gives empirical relative error about 1%
    val minimumPPR = 0.1f / graph.nodeCount
    def estimate(t: Int) = {
      val walkCount = (chernoffMultiplier * reverseThreshold / minimumPPR).toInt
      val (estimates, residuals) = precomputation.

      var estimate = estimates(sourceId)
      for (walkIndex <- 0 until walkCount) {
        val v = samplePPR(sourceId)
        estimate += residuals(v) / walkCount
      }
      estimate
    }
    new mutable.HashMap() ++= (targets map { t => (t, estimate(t))})
    def estimatePPR(t: Int): Float = {
      val chernoffConstant: Int = 100 // This is greater than 12 * e / ln(2 / 0.1) so it guarantees
      // the given relative error with probability 90%.  TODO: update after experiments
      val chernoffMultiplier = chernoffConstant / math.pow(relativeError, 2.0).toFloat
      // TODO: Add balance constant and tune in experiments, or implement balanced version

    }
  }
*/



  def exportReverseMatrix(targetSetSize: Int, graphName: String): Unit = {
    val samplerWalkCount = 100 * 1000
    val powerLawAlpha = 0.77
    val targetK = 3
    // val minPPR = 1.0f / n
    val a = powerLawAlpha
    val n = graph.nodeCount
    val minPPR = ((1.0 - a) / math.pow(targetSetSize, 1.0 - a) * math.pow(targetK, -a) * targetSetSize / n).toFloat
    val samplerChernoffConstant = 20
    val maxResidual = (samplerWalkCount * minPPR / samplerChernoffConstant).toFloat
    val residualIdOffset = 42 * 1000 * 1000

    val targetIds = ArrayBuffer.fill(targetSetSize)(randomNode().id)

    val f = new File(s"$graphName-residuals-${targetSetSize}targets-" +
      s"time${System.currentTimeMillis}-random${random.nextInt()}")
    println("writing estimates and residuals to " + f.getAbsolutePath)
    val out = new BufferedWriter(new FileWriter(f))
    ExperimentUtils.printTime("residual computation and file writing took ") {
      for (t <- targetIds) {
        val (estimates, residuals) = pprEstimator.computeContributions(t, maxResidual)
        for ((v, value) <- estimates) {
          out.write(s"$t $v $value\n")
        }
        for ((v, value) <- residuals) {
          out.write(s"$t ${v + residualIdOffset} $value\n")
        }
      }
    }
    out.close()
  }


  def runtimeTrial(targetSetSize: Int): RuntimeAndAccuracyResult = {
    val startNodeCount = 10
    val n = graph.nodeCount
    val samplerWalkCount = 100 * 1000 // was 10K in v19, 100K in v20
    val samplerChernoffConstant = 20
    val mcChernoffConstant = 40 // Was 20 in v19
    val biPPRRelativeError = math.sqrt(0.07 / samplerChernoffConstant).toFloat
    val maxMCWalkCount = 10 * 1000 * 1000
    val maxPPREstimateCount = 300
    val powerLawAlpha = 0.77
    val targetK = 3
    // val minPPR = 1.0f / n
    val a = powerLawAlpha
    val minPPR = ((1.0 - a) / math.pow(targetSetSize, 1.0 - a) * math.pow(targetK, -a) * targetSetSize / n).toFloat
    //val maxResidual = math.sqrt(minPPR).toFloat
    val maxResidual = (samplerWalkCount * minPPR / samplerChernoffConstant).toFloat
    //val walkCount = (8 *  maxResidual / minPPR).toInt
    val mcWalkCount = (mcChernoffConstant / minPPR).toInt
    println(s"runtimeExperiment using minPPR=$minPPR = ${minPPR * n}/n")
    println(s"samplerWalkCount = $samplerWalkCount")
    println(s"reverseThreshold = $maxResidual")
    println("precomputing...")
    val targets = ArrayBuffer.fill(targetSetSize)(randomNode().id)
    val startNodes = ArrayBuffer.fill(startNodeCount) (randomNonIsolatedStart())

    val precomputation = ExperimentUtils.printTime("sampler precomputation took ") {
      pprSearcher.precomputeForTargetSet(targets, maxResidual)
    }
    println("samplerPrecomputationSize = " + precomputation.size)

    val estimateResidualSeqs = ExperimentUtils.printTime("biPPR precomputation took ") {
      targets map { target =>
        pprEstimator.computeContributions(target, maxResidual)
      }
    }

    val estimatesByV = ExperimentUtils.printTime("biPPR grouping estimates took ") {
      groupByV(targets, estimateResidualSeqs map (_._1))
    }
    val residualsByV = ExperimentUtils.printTime("biPPR grouping residuals took ") {
      groupByV(targets, estimateResidualSeqs map (_._2))
    }

    val (samplingWeights, samplingRunTimes) = (startNodes map { start =>
      ExperimentUtils.time {
        pprSearcher.sampleTargets(start, precomputation, samplerWalkCount, samplerWalkCount)
      }
    }).unzip
    val samplerTime = ExperimentUtils.median(samplingRunTimes)
    println(s"samplerRunTimes: $samplingRunTimes")

    // TODO: Are random walk time estimates fair, given that edges near start may get cached by CPU?
    val (biPPRGroupedWeights, biPPRGroupedTimes) =
      (startNodes map { start =>
        ExperimentUtils.time {
        val forwardVector = pprPrecomputor.forwardPrecomputation(start, minPPR, maxResidual,
          relativeError = biPPRRelativeError)
        computeBiPPRGroupedTargetWeights(start, forwardVector, estimatesByV, residualsByV)
      }
    }).unzip
    val biPPRGroupedTime = ExperimentUtils.median(biPPRGroupedTimes)
    println(s"biPPRGroupedTime: $biPPRGroupedTime = median($biPPRGroupedTimes")


    val biPPRWithoutPrecomputeTimes = startNodes map { start =>
      ExperimentUtils.time {
        for (t <- targets take maxPPREstimateCount) yield {
          pprEstimator.estimatePPR(start, t, minPPR, biPPRRelativeError)
        }
      }._2
    }
    val correctionFactor = math.max(targets.size, maxPPREstimateCount).toFloat / maxPPREstimateCount
    val biPPRWithoutPrecomputeTime = ExperimentUtils.median(biPPRWithoutPrecomputeTimes) * correctionFactor
    println(s"biPPRWithoutPrecomputeTime = $biPPRWithoutPrecomputeTime = $correctionFactor * median($biPPRWithoutPrecomputeTimes)")

    val (biPPRWeights, biPPRTimes) =
      (startNodes map { start =>
        ExperimentUtils.time {
          val forwardVector = pprPrecomputor.forwardPrecomputation(start, minPPR, maxResidual,
            relativeError = 0.05f)
          val estimates = estimateResidualSeqs map { case (estimates, residuals) =>
            estimates(start) + CollectionsUtil.dotMaps(forwardVector, residuals)
          }
          (targets zip estimates).toMap
        }
      }).unzip
    val biPPRTime = ExperimentUtils.median(biPPRTimes)
    println(s"biPPRTime: $biPPRTime = median($biPPRTimes)")
    println("Sanity check: " + (targets take 10 map {t =>
      (biPPRWeights(0)(t), biPPRGroupedWeights(0)(t))
    }).mkString("\n  "))

    val targetSet = new mutable.HashSet() ++= targets
    val (mcTargetWeights, mcTimes) = (startNodes map { start =>
      ExperimentUtils.time {
        estimateMontoCarloTargetWeights(start, targetSet, math.min(maxMCWalkCount, mcWalkCount))
      }
    }).unzip
    println(s"rawMCTimes = $mcTimes")
    val mcTime = if (mcWalkCount <= maxMCWalkCount) {
      ExperimentUtils.median(mcTimes)
    } else {
      println(s"Note: mcWalkCount = $mcWalkCount; multiplying by " + (mcWalkCount.toDouble / maxMCWalkCount))
      ExperimentUtils.median(mcTimes) * mcWalkCount / maxMCWalkCount
    }
    println(s"mcTime = $mcTime")



    val biPPRPrecisionAtKSums = ArrayBuffer.fill(maxK + 1)(0.0)
    val samplingPrecisionAtKSums = ArrayBuffer.fill(maxK + 1)(0.0)
    val mcPrecisionAtKSums = ArrayBuffer.fill(maxK + 1)(0.0)

    for (i <- 0 until startNodes.size) {
      val candidateCount = 10 * maxK
      val candidateTargets = (targets sortBy biPPRWeights(i)).reverse take candidateCount
      //val trueTargetPPRs = estimateTrueTargetPPRs(startNodes(i), candidateTargets)
      val trueTargetPPRs = ExperimentUtils.printTime("computing ground truth took ") {
        val forwardVector = pprPrecomputor.forwardPrecomputation(startNodes(i), minPPR, maxResidual,
          relativeError = 0.01f)
        computeBiPPRGroupedTargetWeights(startNodes(i), forwardVector, estimatesByV, residualsByV)
      }


      val (samplingPrecisions, samplingTotalPPRs) = measurePrecisionAndTotalPPRAtKs(
        targets, samplingWeights(i), trueTargetPPRs, maxK)
      CollectionsUtil.plusEquals(samplingPrecisionAtKSums, samplingPrecisions)

      val (biPPRPrecisions, biPPRTotalPPRs) = measurePrecisionAndTotalPPRAtKs(
        targets, biPPRWeights(i), trueTargetPPRs, maxK)
      CollectionsUtil.plusEquals(biPPRPrecisionAtKSums, biPPRPrecisions)


      // Because we don't actually do the desired number of Monte Carlo walks to save time, simulate MC walks.
      val targetDistribution = DiscreteAliasSampler(trueTargetPPRs, random)
      val piST = trueTargetPPRs.values.sum
      val relevantSampleCount = (mcWalkCount * piST).toInt
      println(s"Sampling $relevantSampleCount times to simulate $mcWalkCount walks, based on pi_s(T)= $piST = ${piST * n} / n")
      val simulatedMCWeights = CollectionsUtil.efficientIntFloatMapWithDefault0()
      for (sampleI <- 0 until relevantSampleCount)
        simulatedMCWeights(targetDistribution.sample()) += 1

      val (mcPrecisions, mcTotalPPRs) = measurePrecisionAndTotalPPRAtKs(
        targets, /*mcTargetWeights(i)*/ simulatedMCWeights, trueTargetPPRs, maxK)
      CollectionsUtil.plusEquals(mcPrecisionAtKSums, mcPrecisions)
    }


    new RuntimeAndAccuracyResult(
      CollectionsUtil.scaleSeq(biPPRPrecisionAtKSums, 1.0 / startNodeCount),
      CollectionsUtil.scaleSeq(samplingPrecisionAtKSums, 1.0 / startNodeCount),
      CollectionsUtil.scaleSeq(mcPrecisionAtKSums, 1.0 / startNodeCount),
      biPPRWithoutPrecomputeTime,
      biPPRTime, biPPRGroupedTime, samplerTime, mcTime, precomputation.size)
  }
/*
  def computeGroupedEstimatesAndResiduals(maxResidual: Float, targets: Seq[Int]
      ): (Map[Int, Map[Int, Float]], Map[Int, Map[Int, Float]]) = {
    val estimateMap = new mutable.AnyRefMap[Int, mutable.Map[Int, Float]]()
    val residualMap = new mutable.AnyRefMap[Int, mutable.Map[Int, Float]]()
    for (t <- targets) {
      val (tEstimates, tResiduals) = pprEstimator.computeContributions(t, maxResidual)
      for ((v, estimate) <- tEstimates) {
        val estimatesByV = estimateMap.getOrElseUpdate(v, CollectionsUtil.efficientIntFloatMapWithDefault0())
        estimatesByV(t) = estimate
      }
      for ((v, residual) <- tResiduals) {
        val residualsByV = estimateMap.getOrElseUpdate(v, CollectionsUtil.efficientIntFloatMapWithDefault0())
        residualsByV(t) = residual
      }
    }
    (estimateMap, residualMap)
  }
  */
  def groupByV(targets: Seq[Int], valueMaps: Seq[Map[Int, Float]]): Map[Int, Map[Int, Float]] = {
    val result = new mutable.HashMap[Int, mutable.Map[Int, Float]]()
    for ((t, tValues) <- targets zip valueMaps) {
      for ((v, value) <- tValues) {
        val resultV = result.getOrElseUpdate(v, CollectionsUtil.efficientIntFloatMapWithDefault0())
        resultV(t) = value
      }
    }
    result
  }

  def computeBiPPRGroupedTargetWeights(
      start: Int,
      forwardVector: Map[Int, Float],
      estimatesByV: Map[Int, Map[Int, Float]],
      residualsByV: Map[Int, Map[Int, Float]]): Map[Int, Float] = {
    val weights = CollectionsUtil.efficientIntFloatMapWithDefault0()
    for ((t, estimate) <- estimatesByV.getOrElse(start, Map.empty)) {
      weights(t) = estimate
    }
    for ((v, pprSV) <- forwardVector) {
      for ((t, residualVT) <- residualsByV.getOrElse(v, Map.empty)) {
        weights(t) += pprSV * residualVT
      }
    }
    weights
  }

}

object PPRSearchExperiments {
  /*
  def approximateSamplingAccuracyExperiment(
      graph: DirectedGraph[Node], teleportProb: Float,
      targetSetSize: Int, trialCount: Int): Unit = {
    println(s"targetSetSize: $targetSetSize")
    val startNodeCount = 10
    val n = graph.nodeCount
    val minPPR = 1.0f / n
    val reverseThreshold = math.sqrt(minPPR).toFloat
    val walkCounts = (0 to 3) map { i => (math.pow(2, i) *  reverseThreshold / minPPR).toInt }
    println(s"walkCounts = " + ExperimentUtils.toPythonList(walkCounts))

    val random = new Random()
    def randomTargetSet() = (0 until targetSetSize) map { i => randomNode().id }

    val results = ((0 until trialCount).par map { trialI =>
      val targets = randomTargetSet()
      // Create a separate object per thread, so prevent any data synchronization issues
      val searchExperiment = new PPRSearchExperiments(graph, teleportProb)
      searchExperiment.measureAccuracyToTargets(
        targets, minPPR, reverseThreshold, walkCounts, startNodeCount)
    }).seq

    val allTotalTargetPPRs = (results map { _.allTotalTargetPPRs}).flatten
    def precisionAt3(precisions: Seq[Double]) = precisions(3)
    val allPrecisionAt3s = (results map { _.samplingPrecisionAtKs map precisionAt3})
    println("allTotalTargetPPRs = " + ExperimentUtils.toPythonList(allTotalTargetPPRs))
    println("allPrecisionAt3s = " + ExperimentUtils.matrixToPythonList(allPrecisionAt3s))

    println(s"walkCounts = " + ExperimentUtils.toPythonList(walkCounts))
    def printMean(label: String, precisionsMatrix: Seq[Seq[Seq[Double]]]) {
      val meanPrecisions = ExperimentUtils.meanMatrix(precisionsMatrix)
      println(s"$label = " + ExperimentUtils.matrixToPythonList(meanPrecisions))
    }
    printMean("meanBiPPRPrecision", results map { _.biPPRPrecisionAtKs})
    printMean("meanSamplingPrecision", results map { _.samplingPrecisionAtKs})
    printMean("meanMCPrecision", results map { _.mcPrecisionAtKs})

    val meanTotalPPRs = ExperimentUtils.meanMatrix(results.seq map { _.samplingTotalPPRAtKs})
    println(s"meanSamplingTotalPPRs = " + ExperimentUtils.matrixToPythonList(meanTotalPPRs))

  }
  */

  def exportReverseMatrixExperiment(graph: DirectedGraph[Node],
                                    graphName: String,
                                    teleportProb: Float,
                                    trialCount: Int,
                                    targetSetSizes: Seq[Int]
                                     ): Unit = {
    for (targetSetSize <- targetSetSizes) {
      (0 until trialCount).par foreach { trialI =>
        // Create a separate object per thread, so prevent any data synchronization issues
        val searchExperiment = new PPRSearchExperiments(graph, teleportProb)
        searchExperiment.exportReverseMatrix(targetSetSize, graphName)
      }
    }
  }

  def runtimeAndAccuracyExperiment(
    graph: DirectedGraph[Node],
    teleportProb: Float,
    trialCount: Int, targetSetSizes: Seq[Int]
    ): Unit = {
    val results = new ArrayBuffer[RuntimeAndAccuracyResult]()
    for (targetSetSize <- targetSetSizes) {
      results += RuntimeAndAccuracyResult.median(
        ((0 until trialCount).par map { trialI =>
          // Create a separate object per thread, so prevent any data synchronization issues
          val searchExperiment = new PPRSearchExperiments(graph, teleportProb)
          searchExperiment.runtimeTrial(targetSetSize)
        }).seq
      )
      println("ResultUpdate =================================")
      println("targetSetSizes = " + ExperimentUtils.toPythonList(targetSetSizes))
      println("precomputationSizes = " + ExperimentUtils.toPythonList(results map (_.samplingPrecomputationSize)))
      println("biPPRWithoutPrecomputeTimes = " + ExperimentUtils.toPythonList(results map {_.biPPRWithoutPrecomputeRunningTime}))
      println("biPPRTimes = " + ExperimentUtils.toPythonList(results map {_.biPPRRunningTime}))
      println("biPPRGroupedTimes = " + ExperimentUtils.toPythonList(results map {_.biPPRGroupedRunningTime}))
      println("samplingTimes = " + ExperimentUtils.toPythonList(results map {_.samplingRunningTime}))
      println("mcTimes = " + ExperimentUtils.toPythonList(results map {_.mcRunningTime}))
      println("biPPRPrecisions = " + ExperimentUtils.matrixToPythonList(results map (_.biPPRPrecisionAtKs)))
      println("samplingPrecisions = " + ExperimentUtils.matrixToPythonList(results map (_.samplingPrecisionAtKs)))
      println("mcPrecisions = " + ExperimentUtils.matrixToPythonList(results map (_.mcPrecisionAtKs)))
      println("biPPRPrecisionAt3s = " + ExperimentUtils.toPythonList(results map (_.biPPRPrecisionAtKs(2))))
      println("samplingPrecisionAt3s = " + ExperimentUtils.toPythonList(results map (_.samplingPrecisionAtKs(2))))
      println("mcPrecisionAt3s = " + ExperimentUtils.toPythonList(results map (_.mcPrecisionAtKs(2))))
    }
  }
}

case class AccuracyResult(biPPRPrecisionAtKs: Seq[Seq[Double]],
                          samplingPrecisionAtKs: Seq[Seq[Double]],
                          mcPrecisionAtKs: Seq[Seq[Double]],
                          samplingTotalPPRAtKs: Seq[Seq[Double]],
                          allTotalTargetPPRs: Seq[Double])

case class RuntimeAndAccuracyResult(
  biPPRPrecisionAtKs: Seq[Double],
  samplingPrecisionAtKs: Seq[Double],
  mcPrecisionAtKs: Seq[Double],
  biPPRWithoutPrecomputeRunningTime: Double,
  biPPRRunningTime: Double,
  biPPRGroupedRunningTime: Double,
  samplingRunningTime: Double,
  mcRunningTime: Double,
  samplingPrecomputationSize: Double)

object RuntimeAndAccuracyResult {
  def mean(results: Seq[RuntimeAndAccuracyResult]): RuntimeAndAccuracyResult =
    RuntimeAndAccuracyResult(
      ExperimentUtils.meanSeq(results map (_.biPPRPrecisionAtKs)),
      ExperimentUtils.meanSeq(results map (_.samplingPrecisionAtKs)),
      ExperimentUtils.meanSeq(results map (_.mcPrecisionAtKs)),
      ExperimentUtils.mean(results map (_.biPPRWithoutPrecomputeRunningTime)),
      ExperimentUtils.mean(results map (_.biPPRRunningTime)),
      ExperimentUtils.mean(results map (_.biPPRGroupedRunningTime)),
      ExperimentUtils.mean(results map (_.samplingRunningTime)),
      ExperimentUtils.mean(results map (_.mcRunningTime)),
      ExperimentUtils.mean(results map (_.samplingPrecomputationSize))
    )
  def median(results: Seq[RuntimeAndAccuracyResult]): RuntimeAndAccuracyResult =
    RuntimeAndAccuracyResult(
      ExperimentUtils.medianSeq(results map (_.biPPRPrecisionAtKs)),
      ExperimentUtils.medianSeq(results map (_.samplingPrecisionAtKs)),
      ExperimentUtils.medianSeq(results map (_.mcPrecisionAtKs)),
      ExperimentUtils.median(results map (_.biPPRWithoutPrecomputeRunningTime)),
      ExperimentUtils.median(results map (_.biPPRRunningTime)),
      ExperimentUtils.median(results map (_.biPPRGroupedRunningTime)),
      ExperimentUtils.median(results map (_.samplingRunningTime)),
      ExperimentUtils.median(results map (_.mcRunningTime)),
      ExperimentUtils.median(results map (_.samplingPrecomputationSize))
    )
}