package soal.ppr.experiments

import com.twitter.cassovary.graph.{Node, DirectedGraph}
import scala.collection.mutable
import soal.ppr.BidirectionalPPREstimator

class RuntimeExperiments(graph: DirectedGraph[Node], teleportProb: Float) {
  val pprEstimator = new BidirectionalPPREstimator(graph, teleportProb)

  def forwardReverseTimesExperiment(trialCount: Int, msPerWalks: Seq[Float]): Unit = {
    val forwardReverseTimes = msPerWalks map {measureMeanForwardReverseTime(trialCount, _)}
    println("msPerWalks " + ExperimentUtils.toPythonList(msPerWalks))
    println("forwardTimes " + ExperimentUtils.toPythonList(forwardReverseTimes map {_._1}))
    println("reverseTimes " + ExperimentUtils.toPythonList(forwardReverseTimes map {_._2}))
  }

  def measureMeanForwardReverseTime(trialCount: Int, msPerWalk: Float): (Double, Double) = {

    val forwardTimes = mutable.ArrayBuffer[Float]()
    val reverseTimes = mutable.ArrayBuffer[Float]()
    for (trialIndex <- 0 until trialCount) {
      val start = ExperimentUtils.randomNode(graph)
      val target = ExperimentUtils.sampleFromPageRank(graph, teleportProb)
      val (forwardTime, reverseTime) = measureForwardReverseTime(start.id, target.id, msPerWalk)
      forwardTimes.append(forwardTime)
      reverseTimes.append(reverseTime)
    }

    val totalTimes = (forwardTimes zip reverseTimes) map { case (x, y) => x + y }
    val meanForwardTime = ExperimentUtils.mean(forwardTimes)
    val meanReverseTime = ExperimentUtils.mean(reverseTimes)
    println("reverseTimes = " + reverseTimes.mkString("[", ", ", "]"))
    println("forwardTimes = " + forwardTimes.mkString("[", ", ", "]"))
    println("totalTimes = " + totalTimes.mkString("[", ", ", "]"))
    println("meanForwardTime = " + meanForwardTime)
    println("meanReverseTime = " + meanReverseTime)
    /*if (balanced) {
      println("Current forwardStepsPerReverseStep value = " + config.forwardStepsPerReverseStep)
      println("Suggested new forwardStepsPerReverseStep value = " +
        config.forwardStepsPerReverseStep * math.sqrt(reverseTimes.sum / forwardTimes.sum))
    }
    */
    (meanForwardTime, meanReverseTime)
  }

  // From http://rosettacode.org/wiki/Time_a_function#Scala
  def time(f: => Unit): Long = {
    val s = System.currentTimeMillis
    f
    System.currentTimeMillis - s
  }

  def measureForwardReverseTime(startId: Int,
                                targetId: Int,
                                 msPerWalk: Float): (Float, Float) = {
    // This is copied and pasted from FastPPR. TODO: find a way of measure forward+backward time without copying+pasting
    val minimumPPR = 4.0f / graph.nodeCount
    val chernoffMultiplier = 7.0f

    def computeWalkCount(maxResidual: Float): Int =
      (chernoffMultiplier * maxResidual / minimumPPR).toInt
    def estimateForwardTimeInMillis(maxResidual: Float): Float =
      computeWalkCount(maxResidual) * msPerWalk


    val ((estimates, residuals, maxResidual), reverseTime) = ExperimentUtils.time {
      pprEstimator.computeContributionsBalanced(targetId, estimateForwardTimeInMillis)
    }

    val walkCount = computeWalkCount(maxResidual)
    val (_, forwardTime) = ExperimentUtils.time {
      var estimate = estimates(startId)
      for (walkIndex <- 0 until walkCount) {
        val v = pprEstimator.samplePPR(startId)
        estimate += residuals(v) / walkCount
      }
    }
    println(s"walkCount reverseSize forwardTime reverseTime $walkCount ${residuals.size + estimates.size} $forwardTime $reverseTime" )
    (forwardTime.toFloat, reverseTime.toFloat)
  }
}
