package soal.ppr

import org.scalatest.{Matchers, FlatSpec}
import java.util.Random
import soal.util.CassovaryUtil

class BidirectionalPPRSearcherSpec extends FlatSpec with Matchers {
  val graph = CassovaryUtil.readGraph("src/test/resources/test_graph.txt")
  val truePPRs = BidirectionalPPREstimatorSpec.testGraphTruePPRs
  val teleportProb = 0.2f
  val random = new Random(2) // Seed for consistent tests
  val pprSearcher = new BidirectionalPPRSearcher(graph, teleportProb, random)

  "A BidirectionalPPRSearcher" should "sample correctly" in {
    val targetSet = Array(1, 3)
    // maxResidual is chosen to contribution of estimates from 0 to targets is comparable to
    // contribution of residuals
    val maxResidual = 0.2f
    val tol = 0.01f
    val walkCount = 10000
    val sampleCount = walkCount

    val precomputation = pprSearcher.precomputeForTargetSet(targetSet, maxResidual)


    val targetWeightsFrom0 = pprSearcher.sampleTargets(0, precomputation,
      walkCount, sampleCount)
    targetWeightsFrom0(1) should equal (truePPRs((0, 1)) +- tol)
    val targetSetPPR = (targetSet map { t => truePPRs((0, t)) }).sum
    val trueFraction0to1 = truePPRs((0, 1)) / targetSetPPR
    val actualFraction0to1 = targetWeightsFrom0(1) / targetWeightsFrom0.values.sum
    actualFraction0to1 should equal (trueFraction0to1 +- tol)
  }

  it should "sample correctly with a size 4 target set" in {
    val targetSet = Array(0, 1, 3, 4).toSeq
    // maxResidual is chosen to contribution of estimates from 0 to targets (pSampleEstimate) is
    // comparable to
    // contribution of residuals TODO
    val maxResidual = 0.2f
    val walkCount = 10000
    val sampleCount = walkCount
    val tol = 0.02f

    val precomputation = pprSearcher.precomputeForTargetSet(targetSet, maxResidual)
    //println(s"precomputation: $precomputation")

    val source = 2
    val targetWeights = pprSearcher.sampleTargets(source, precomputation, walkCount, sampleCount)
    val targetSetPPR = (targetSet map { t => truePPRs((source, t)) }).sum
    for (t <- targetSet) {
      //val trueFraction = truePPRs((source, t)) / targetSetPPR
      withClue(s"testing $source to $t in $targetSet") {
        targetWeights(t) should equal (truePPRs((source, t)) +- tol)
        //(targetWeights(t) / targetWeights.values.sum) should equal (trueFraction +- tol)
      }
    }
  }

  val graph2 = CassovaryUtil.readGraph("src/test/resources/test_graph2.txt")
  val pprSearcher2 = new BidirectionalPPRSearcher(graph2, teleportProb, random)

  it should "not throw exceptions given a disconnected graph" in {
    val targetSet = Array(1, 2)
    // maxResidual is chosen to contribution of estimates from 0 to targets is comparable to
    // contribution of residuals
    val maxResidual = 0.2f
    val walkCount = 10000
    val sampleCount = walkCount

    val precomputation = pprSearcher2.precomputeForTargetSet(targetSet, maxResidual)
    println(precomputation)
    val targetWeightsFrom0 = pprSearcher2.sampleTargets(0, precomputation, walkCount, sampleCount)
    println(targetWeightsFrom0)
    assert(targetWeightsFrom0.isEmpty)
  }
}
